/*----------------------------------------------------------------------
  File    : JMLP.java
  Contents: Java program to visualize multi-layer perceptron training
  Author  : Christian Borgelt
  History : 2023.04.29 file created from JSkel.java
            2023.05.17 refactoring based on BarChart.java
----------------------------------------------------------------------*/
package jmlp;

import java.util.Timer;
import java.util.TimerTask;
import java.io.IOException;
import java.io.File;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Container;
import java.awt.Font;
import java.awt.BorderLayout;
import java.awt.FlowLayout;
import java.awt.GridLayout;
import java.awt.GridBagLayout;
import java.awt.GridBagConstraints;
import java.awt.EventQueue;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import javax.swing.ImageIcon;
import javax.swing.Box;
import javax.swing.BorderFactory;
import javax.swing.JFrame;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JDialog;
import javax.swing.JLabel;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JTextField;
import javax.swing.JTextArea;
import javax.swing.JFileChooser;
import javax.swing.JOptionPane;

/*--------------------------------------------------------------------*/
/** Class for the main window in which the neural network is drawn.
 *  @author Christian Borgelt
 *  @since  2023.04.30 */
/*--------------------------------------------------------------------*/
public class JMLP extends JFrame implements Runnable {

  private static final long serialVersionUID = 0x00010002L;
  /** the version string */
  public  static final String VERSION = "1.2 (2023.05.17)";

  /*------------------------------------------------------------------*/
  /*  constants                                                       */
  /*------------------------------------------------------------------*/
  /** the font for input text fields */
  public  static final Font BOLD = new Font("Dialog", Font.BOLD, 12);

  /*------------------------------------------------------------------*/
  /*  instance variables                                              */
  /*------------------------------------------------------------------*/
  /** the epochs display in menu bar (right-aligned) */
  private JMenu          epochs  = null;

  /** menu item: load weights */
  private JMenuItem      mload   = null;
  /** menu item: save weights */
  private JMenuItem      msave   = null;
  /** menu item: start training */
  private JMenuItem      mstart  = null;
  /** menu item: stop  training */
  private JMenuItem      mstop   = null;
  /** menu item: set parameters */
  private JMenuItem      mparms  = null;
  /** menu item: set weights */
  private JMenuItem      mwgts   = null;

  /** the main panel to draw on */
  private JMLPPanel      panel   = null;
  /** a file chooser (for loading/saving weights) */
  private JFileChooser   chooser = null;

  /** the parameter dialog (e.g. learning rate, momentum factor) */
  private JDialog        params  = null;
  /** the combo box for the training mode (online or batch) */
  private JComboBox      jtmode  = null;
  /** the input field for the learning rate */
  private JTextField     jlrate  = null;
  /** the input field for the momentum factor */
  private JTextField     jmoment = null;
  /** the input field for the initial weight range */
  private JTextField     jrange  = null;
  /** the input field for the delay between steps (in seconds) */
  private JTextField     jdelay  = null;

  /** the connection weights and bias values dialog */
  private JDialog        weights = null;
  /** the input fields for connection weights and bis values */
  private JTextField[][] jwgts   = null;

  /** the "About..." dialog box */
  private JDialog        about   = null;

  /** the timer for scheduling training steps */
  private Timer          timer   = new Timer();
  /** the training task for scheduling with a timer */
  private TimerTask      task    = null;
  /** the delay between training steps (in seconds) */
  private double         delay   = 0.01;

  /*------------------------------------------------------------------*/
  /** Create a main window.
   *  @since  2023.04.30 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public JMLP ()
  {                             /* --- constructor of JMLP class */
    if (EventQueue.isDispatchThread()) { this.run(); return; }
    try { EventQueue.invokeAndWait(this); }
    catch (Exception e) {}      /* create the user interface */
  }  /* JLVQ() */

  /*------------------------------------------------------------------*/
  /** Create the user interface.
   *  <p>Following the recommendations in the Java tutorial, the user
   *  interface is created in the "run" method, which is invoked from
   *  the event queue, in order to avoid problems with threads.</p>
   *  @since  2023.05.17 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void run ()
  {                             /* --- create an MLP visualization */
    JMenuBar  mbar;             /* menu bar */
    JMenu     menu;             /* to create menu titles */
    JMenuItem item;             /* to create menu items */

    /* --- create and set the menu bar --- */
    this.setJMenuBar(mbar = new JMenuBar());

    menu = mbar.add(new JMenu("File"));
    menu.setMnemonic('f');
    this.mload = item = menu.add(new JMenuItem("Load Weights..."));
    item.setMnemonic('l');
    item.addActionListener(new ActionListener() {
      public void actionPerformed (ActionEvent e) {
        JMLP.this.loadWeights(null); } } );
    this.msave = item = menu.add(new JMenuItem("Save Weights..."));
    item.setMnemonic('s');
    item.addActionListener(new ActionListener() {
      public void actionPerformed (ActionEvent e) {
        JMLP.this.saveWeights(null); } } );
    menu.addSeparator();
    item = menu.add(new JMenuItem("Quit"));
    item.setMnemonic('q');
    item.addActionListener(new ActionListener() {
      public void actionPerformed (ActionEvent e) {
        System.exit(0); } } );

    menu = mbar.add(new JMenu("Actions"));
    menu.setMnemonic('a');
    item = menu.add(new JMenuItem("Exclusive Or"));
    item.setMnemonic('e');
    item.addActionListener(new ActionListener() {
      public void actionPerformed (ActionEvent e) {
        JMLP.this.panel.setType(MLP.XOR); } });
    item = menu.add(new JMenuItem("Biimplication"));
    item.setMnemonic('b');
    item.addActionListener(new ActionListener() {
      public void actionPerformed (ActionEvent e) {
        JMLP.this.panel.setType(MLP.BIIMP); } });
    item = menu.add(new JMenuItem("Function 1"));
    item.setMnemonic('1');
    item.addActionListener(new ActionListener() {
      public void actionPerformed (ActionEvent e) {
        JMLP.this.panel.setType(MLP.FUNC1); } });
    item = menu.add(new JMenuItem("Function 2"));
    item.setMnemonic('2');
    item.addActionListener(new ActionListener() {
      public void actionPerformed (ActionEvent e) {
        JMLP.this.panel.setType(MLP.FUNC2); } });
    menu.addSeparator();
    this.mstart = item = menu.add(new JMenuItem("Start Training"));
    item.setMnemonic('x');
    item.addActionListener(new ActionListener() {
      public void actionPerformed (ActionEvent e) {
        JMLP.this.start(); } } );
    this.mstop = item = menu.add(new JMenuItem("Stop Training"));
    item.setMnemonic('x');
    item.setEnabled(false);
    item.setVisible(false);
    item.addActionListener(new ActionListener() {
      public void actionPerformed (ActionEvent e) {
        JMLP.this.stop(); } });
    menu.addSeparator();
    item = menu.add(new JMenuItem("Redraw"));
    item.setMnemonic('r');
    item.addActionListener(new ActionListener() {
      public void actionPerformed (ActionEvent e) {
        JMLP.this.panel.repaint(); } });

    menu = mbar.add(new JMenu("Settings"));
    menu.setMnemonic('s');
    this.mparms = item = menu.add(new JMenuItem("Parameters..."));
    item.setMnemonic('p');
    item.addActionListener(new ActionListener() {
      public void actionPerformed (ActionEvent e) {
        JMLP.this.getParamsDlg().setVisible(true); } });
    this.mwgts = item = menu.add(new JMenuItem("Weights..."));
    item.setMnemonic('w');
    item.addActionListener(new ActionListener() {
      public void actionPerformed (ActionEvent e) {
        JMLP.this.getWeightsDlg().setVisible(true); } });

    menu = mbar.add(new JMenu("Help"));
    menu.setMnemonic('h');
    item = menu.add(new JMenuItem("About..."));
    item.setMnemonic('a');
    item.addActionListener(new ActionListener() {
      public void actionPerformed (ActionEvent e) {
        JMLP.this.getAboutDlg().setVisible(true); } });
    mbar.add(Box.createHorizontalGlue());
    this.epochs = mbar.add(new JMenu("0"));
    
    /* --- create and set the main panel --- */
    this.panel = new JMLPPanel(MLP.XOR);
    this.panel.setLayout(new BorderLayout());
    this.panel.setPreferredSize(new Dimension(572, 400));
    this.add(panel, BorderLayout.CENTER);

    /* --- create and show the frame window --- */
    this.setTitle("JMLP");
    this.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
    this.setLocation(48, 48);
    this.setResizable(true);
    this.pack();
    this.setVisible(true);
  }  /* JMLP() */

  /*------------------------------------------------------------------*/
  /** Create a file chooser (for saving / loading weights).
   *  @since  2023.04.29 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  private JFileChooser getFileChooser ()
  {                             /* --- get the file chooser */
    if (this.chooser != null)   /* if the chooser already exists, */
      return this.chooser;      /* simply return it */
    this.chooser = new JFileChooser();
    this.chooser.setFileSelectionMode(JFileChooser.FILES_ONLY);
    this.chooser.setCurrentDirectory(new File("."));
    this.chooser.setFileHidingEnabled(true);
    this.chooser.setAcceptAllFileFilterUsed(true);
    this.chooser.setMultiSelectionEnabled(false);
    this.chooser.setFileView(null);
    return this.chooser;        /* create and configure the chooser */
  }  /* getFileChooser() */

  /*------------------------------------------------------------------*/
  /** Load neural network weights.
   *  @param  file  the file to read the weights from
   *  @return whether loading the network was successful
   *  @since  2023.04.30 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public boolean loadWeights (File file)
  {                             /* --- load a pointgon */
    if (file == null) {         /* if no file name is given */
      int r = this.getFileChooser().showOpenDialog(this);
      if (r != JFileChooser.APPROVE_OPTION) return false;
      file = this.chooser.getSelectedFile();
    }                           /* get the selected file */
    try {                       /* load the weights */
      this.panel.getMLP().loadWeights(file); }
    catch (IOException e) {     /* catch and report i/o errors */
      String msg = "file " +file.getName() +":\n" +e.getMessage();
      System.err.println(msg);
      JOptionPane.showMessageDialog(this, msg,
        "Error", JOptionPane.ERROR_MESSAGE);
    }                           /* set the status text */
    this.repaint();             /* repaint the panel */
    return true;                /* return 'success' */
  }  /* loadWeights() */

  /*------------------------------------------------------------------*/
  /** Save neural network weights.
   *  @param  file  the file to write the weights to
   *  @return whether saving the weights was successful
   *  @since  2023.04.30 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public boolean saveWeights (File file)
  {                             /* --- save a pointgon */
    if (file == null) {         /* if no file name is given */
      int r = this.getFileChooser().showSaveDialog(this);
      if (r != JFileChooser.APPROVE_OPTION) return false;
      file = this.chooser.getSelectedFile();
    }                           /* get the selected file */
    try {                       /* create a writer for the file */
      this.panel.getMLP().saveWeights(file); }
    catch (IOException e) {     /* catch and report i/o errors */
      String msg = "file " +file.getName() +":\n" +e.getMessage();
      System.err.println(msg);
      JOptionPane.showMessageDialog(this, msg,
        "Error", JOptionPane.ERROR_MESSAGE);
    }                           /* set the status text */
    return true;                /* return 'success' */
  }  /* saveWeights() */

  /*------------------------------------------------------------------*/
  /** Format a floating point number.
   *  @param  x  the number to be formated
   *  @return a string containing the formatted number
   *  @since  2023.05.01 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  static private String format (double x)
  {                             /* --- format a floating point number */
    int    i;                   /* loop variable */
    String s;                   /* initially formatted number */

    s = String.format("%g", x); /* do initial string formatting */
    for (i = s.length(); --i >= 0; )  /* remove all */
      if (s.charAt(i) != '0') break;  /* trailing zeros */
    if (i < 0) i = 0;           /* maintain at least one character */
    if (s.charAt(i) == '.') i -= 1;   /* remove decimal point */
    return s.substring(0,i+1);  /* return the formatted number */
  }  /* format() */

  /*------------------------------------------------------------------*/
  /** Set parameters in parameter dialog box.
   *  @since  2023.04.30 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void params2dlg ()
  {                             /* --- set parameters in dialog */
    JMLPPanel pnl = this.panel; /* panel and multi-layer perceptron */
    MLP       mlp = pnl.getMLP();

    this.jtmode.setSelectedIndex(mlp.getMode());
    this.jlrate.setText(format(mlp.getLRate()));
    this.jmoment.setText(format(mlp.getMoment()));
    this.jrange.setText(format(pnl.getRange()));
    this.jdelay.setText(format(this.delay));
  }  /* params2dlg() */

  /*------------------------------------------------------------------*/
  /** Save neural network weights.
   *  @since  2023.04.30 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void dlg2params ()
  {                             /* --- get parameters from dialog */
    JMLPPanel pnl = this.panel; /* panel and multi-layer perceptron */
    MLP       mlp = pnl.getMLP();

    mlp.setMode(this.jtmode.getSelectedIndex());
    try { mlp.setLRate(Double.parseDouble(this.jlrate.getText())); }
    catch (Exception x) { mlp.setLRate(0.2); }
    try { mlp.setMoment(Double.parseDouble(this.jmoment.getText())); }
    catch (Exception x) { mlp.setMoment(0.0); }
    try { pnl.setRange(Double.parseDouble(this.jrange.getText())); }
    catch (Exception x) { pnl.setRange(1.0); }
    try { this.delay = Double.parseDouble(this.jdelay.getText()); }
    catch (Exception x) { this.delay = 0.01; }
    if (this.delay < 0.001) this.delay = 0.001;
  }  /* dlg2params() */

  /*------------------------------------------------------------------*/
  /** Create parameter dialog (learning rate, momentum factor etc.).
   *  @return the created parameter dialog
   *  @since  2023.04.30 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  private JDialog getParamsDlg ()
  {                             /* --- create parameters dialog */
    if (this.params != null) {  /* if dialog exists, return it */
      params2dlg(); return this.params; }

    this.params = new JDialog();
    GridBagLayout      gbl  = new GridBagLayout();
    GridBagConstraints lc   = new GridBagConstraints();
    GridBagConstraints rc   = new GridBagConstraints();
    JPanel             grid = new JPanel(gbl);
    JPanel             bbar = new JPanel(new GridLayout(1, 2, 5, 5));
    JLabel             lbl;     /* swing objects */
    JButton            btn;     /* needed for the dialog */

    grid.setBorder(BorderFactory.createEmptyBorder(5, 5, 5, 5));
    lc.fill      =              /* fill fields in both directions */
    rc.fill      = GridBagConstraints.BOTH;
    rc.weightx   = 1.0;         /* resize only the input fields, */
    lc.weightx   = 0.0;         /* but not the labels */
    lc.weighty   = 1.0;         /* resize lines equally */
    rc.weighty   = 1.0;         /* in vertical direction */
    lc.ipadx     = 10;          /* gap between labels and inputs */
    lc.ipady     = 10;          /* make all lines of the same height */
    rc.gridwidth = GridBagConstraints.REMAINDER;

    lbl = new JLabel("Training mode:");
    gbl.setConstraints(lbl, lc); grid.add(lbl);
    this.jtmode = new JComboBox<String>(new String[]
      { "online", "batch" });
    this.jtmode.setSelectedIndex(1);
    gbl.setConstraints(this.jtmode, rc); grid.add(this.jtmode);

    lbl = new JLabel("Learning rate:");
    gbl.setConstraints(lbl, lc);    grid.add(lbl);
    this.jlrate = new JTextField("0.2");
    this.jlrate.setFont(BOLD);
    gbl.setConstraints(this.jlrate, rc);  grid.add(this.jlrate);

    lbl = new JLabel("Momentum factor:");
    gbl.setConstraints(lbl, lc);    grid.add(lbl);
    this.jmoment = new JTextField("0.0");
    this.jmoment.setFont(BOLD);
    gbl.setConstraints(this.jmoment, rc); grid.add(this.jmoment);

    lbl = new JLabel("Initial weight range:");
    gbl.setConstraints(lbl, lc);    grid.add(lbl);
    this.jrange = new JTextField("1.0");
    this.jrange.setFont(BOLD);
    gbl.setConstraints(this.jrange, rc);  grid.add(this.jrange);

    lbl = new JLabel("Delay (in seconds):");
    gbl.setConstraints(lbl, lc);    grid.add(lbl);
    this.jdelay = new JTextField("0.01");
    this.jdelay.setFont(BOLD);
    gbl.setConstraints(this.jdelay, rc);  grid.add(this.jdelay);

    bbar.setBorder(BorderFactory.createEmptyBorder(5, 0, 0, 0));
    btn = new JButton("Ok"); bbar.add(btn);
    btn.addActionListener(new ActionListener () {
      public void actionPerformed (ActionEvent e) {
        JMLP.this.params.setVisible(false);
	JMLP.this.dlg2params(); } } );
    btn = new JButton("Cancel"); bbar.add(btn);
    btn.addActionListener(new ActionListener () {
      public void actionPerformed (ActionEvent e) {
        JMLP.this.params.setVisible(false); } } );
    rc.ipady = 0;               /* put the 'Ok' and the 'Cancel' */
    gbl.setConstraints(bbar, rc); /* button into a grid layout to make */
    grid.add(bbar);             /* sure that they have the same size */

    this.params.getContentPane().add(grid);
    this.params.setTitle("Parameters...");
    this.params.setModal(true);
    this.params.setLocationRelativeTo(this);
    this.params.pack();
    params2dlg();
    return this.params;
  }  /* getParamsDlg() */

  /*------------------------------------------------------------------*/
  /** Set weights and bias values in weights dialog box.
   *  @since  2023.05.01 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void mlp2dlg ()
  {                             /* --- transfer MLP to dialog */
    MLP mlp = this.panel.getMLP(); /* underlying neural network */
    for (int i = 0; i < 3; i++)
      for (int k = 0; k < 3; k++)
        this.jwgts[i][k].setText(format(mlp.getWeight(i,k)));
  }  /* mlp2dlg() */

  /*------------------------------------------------------------------*/
  /** Get weights and bias values from weights dialog box.
   *  @since  2023.05.01 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void dlg2mlp ()
  {                             /* --- transfer dialog to MLP */
    MLP    mlp = this.panel.getMLP(); /* underlying neural network */
    double w;                   /* buffer for a weight */

    mlp.init(0, null);          /* reinitialize network */
    for (int i = 0; i < 3; i++) {
      for (int k = 0; k < 3; k++) {
        try { w = Double.parseDouble(this.jwgts[i][k].getText()); }
        catch (Exception e) { w = 0.0; }
        mlp.setWeight(i, k, w); /* parse the input value */
      }                         /* and store it in the MLP */
    }
    this.epochs.setText(String.format("%d", mlp.getEpoch()));
  }  /* dlg2mlp() */

  /*------------------------------------------------------------------*/
  /** Create weights dialog (weights and bias values) .
   *  @return the created weights dialog
   *  @since  2023.05.01 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  private JDialog getWeightsDlg ()
  {                             /* --- create parameters dialog */
    if (this.weights != null) { /* if dialog exists, return it */
      mlp2dlg(); return this.weights; }

    this.weights = new JDialog();
    GridBagLayout      gbl  = new GridBagLayout();
    GridBagConstraints lc   = new GridBagConstraints();
    GridBagConstraints mc   = new GridBagConstraints();
    GridBagConstraints rc   = new GridBagConstraints();
    JPanel             grid = new JPanel(gbl);
    JPanel             bbar = new JPanel(new GridLayout(1, 2, 5, 5));
    JLabel             lbl;     /* swing objects */
    JButton            btn;     /* needed for the dialog */
    JTextField         jtf;
    Dimension          d;

    grid.setBorder(BorderFactory.createEmptyBorder(5, 5, 5, 5));
    lc.fill      = mc.fill =    /* fill fields in both directions */
    rc.fill      = GridBagConstraints.BOTH;
    rc.weightx   = 1.0;         /* resize only the input fields, */
    mc.weightx   = 1.0;         /* (middle and right fields), */
    lc.weightx   = 0.0;         /* but not the labels */
    lc.weighty   = 1.0;         /* resize lines equally */
    rc.weighty   = 1.0;         /* in vertical direction */
    lc.ipadx     = 10;          /* gap between labels and inputs */
    lc.ipady     = 10;          /* make all lines of the same height */
    rc.ipadx     = mc.ipadx = rc.ipady = mc.ipady = 1;
    rc.gridwidth = GridBagConstraints.REMAINDER;

    d = new Dimension(100,10);
    this.jwgts   = new JTextField[3][];
    for (int i = 0; i < 3; i++) {
      this.jwgts[i] = new JTextField[3];
      if      (i == 0) lbl = new JLabel("Lower hidden neuron:");
      else if (i == 1) lbl = new JLabel("Upper hidden neuron:");
      else             lbl = new JLabel("Output neuron:");
      gbl.setConstraints(lbl, lc); grid.add(lbl);
      this.jwgts[i][0] = jtf = new JTextField("0.1");
      jtf.setFont(BOLD);
      jtf.setPreferredSize(d); //jtf.setMinimumSize(d);
      gbl.setConstraints(jtf, mc); grid.add(jtf);
      this.jwgts[i][1] = jtf = new JTextField("0.02");
      jtf.setFont(BOLD);
      jtf.setPreferredSize(d); //jtf.setMinimumSize(d);
      gbl.setConstraints(jtf, mc); grid.add(jtf);
      this.jwgts[i][2] = jtf = new JTextField("0.003");
      jtf.setFont(BOLD);
      jtf.setPreferredSize(d); //jtf.setMinimumSize(d);
      gbl.setConstraints(jtf, rc); grid.add(jtf);
    }

    bbar.setBorder(BorderFactory.createEmptyBorder(5, 0, 0, 0));
    btn = new JButton("Ok"); bbar.add(btn);
    btn.addActionListener(new ActionListener () {
      public void actionPerformed (ActionEvent e) {
        JMLP.this.weights.setVisible(false); JMLP.this.dlg2mlp();
        JMLP.this.repaint(); } } );
    btn = new JButton("Cancel"); bbar.add(btn);
    btn.addActionListener(new ActionListener () {
      public void actionPerformed (ActionEvent e) {
        JMLP.this.weights.setVisible(false); } } );
    rc.ipady = 0;               /* put the 'Ok' and the 'Cancel' */
    gbl.setConstraints(bbar, rc); /* button into a grid layout to make */
    grid.add(bbar);             /* sure that they have the same size */

    this.weights.getContentPane().add(grid);
    this.weights.setTitle("Weights...");
    this.weights.setModal(true);
    this.weights.setLocationRelativeTo(this);
    this.weights.pack();
    mlp2dlg();
    return this.weights;
  }  /* getWeightsDlg() */

  /*------------------------------------------------------------------*/
  /** Create an "About" dialog box.
   *  @return the created "About" dialog
   *  @since  2023.04.30 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  private JDialog getAboutDlg ()
  {                             /* --- create "About..." dialog box */
    if (this.about != null)     /* if dialog box already exists, */
      return this.about;        /* simply return it */

    this.about = new JDialog(); /* create dialog box */
    JTextArea text = new JTextArea("JMLP\n"
      +"Multi-layer Perceptron Training Visualization\n"
      +"Version " +JMLP.VERSION +"\n\n"
      +"written by Christian Borgelt\n"
      +"Paris-Lodron-University of Salzburg\n"
      +"Jakob-Haringer-Strasse 2\n"
      +"A-5020 Salzburg, Austria\n"
      +"christian@borgelt.net");
    text.setBackground(this.getBackground());
    text.setFont(new Font("Dialog", Font.BOLD, 12));
    text.setEditable(false);
    JButton btn = new JButton("Ok");
    btn.addActionListener(new ActionListener () {
      public void actionPerformed (ActionEvent e) {
        JMLP.this.about.setVisible(false); } } );
    JPanel rest = new JPanel(new BorderLayout(2, 2));
    rest.setBackground(this.getBackground());
    rest.add(new LogoPanel(), BorderLayout.NORTH);
    rest.add(btn,             BorderLayout.SOUTH);
    Container pane = this.about.getContentPane();
    pane.setLayout(new FlowLayout());
    pane.setBackground(this.getBackground());
    pane.add(text);
    pane.add(rest);
    this.about.setTitle("About JMLP...");
    this.about.setModal(true);
    this.about.setLocationRelativeTo(this);
    this.about.pack();
    this.about.setResizable(false);
    return this.about;
  }  /* getAboutDlg() */

  /*------------------------------------------------------------------*/
  /** Run multi-layer perceptron training.
   *  @since  2023.05.01 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  private void train ()
  {                             /* --- execute one training step */
    MLP mlp = this.panel.getMLP();  /* get the undelying MLP */
    mlp.train();                /* execute one training step */
    this.repaint();             /* repaint the panel */
    this.epochs.setText(String.format("%d", mlp.getEpoch()));
  }  /* train() */

  /*------------------------------------------------------------------*/
  /** Start multi-layer perceptron training.
   *  @since  2023.05.01 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  private void start ()
  {                             /* --- start MLP training */
    this.mstart.setEnabled(false);  /* adapt menu items */
    this.mstart.setVisible(false);  /* so that they show */
    this.mstop.setEnabled(true);    /* the active state */
    this.mstop.setVisible(true);
    this.mparms.setEnabled(false);
    this.mwgts.setEnabled(false);
    this.task = new TimerTask () {  /* create the training task */
      public void run() { JMLP.this.train(); } };
    this.timer.schedule(this.task, 0,
                        Math.max((int)(this.delay *1000), 1));
  }  /* start() */

  /*------------------------------------------------------------------*/
  /** Stop multi-layer perceptron training.
   *  @since  2023.05.01 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  private void stop ()
  {                             /* --- stop MLP training */
    this.task.cancel();         /* cancel the timer task */
    this.mstart.setEnabled(true);   /* adapt menu items */
    this.mstart.setVisible(true);   /* so that they show */
    this.mstop.setEnabled(false);   /* the passive state */
    this.mstop.setVisible(false);
    this.mparms.setEnabled(true);
    this.mwgts.setEnabled(true);
  }  /* stop() */

  /*------------------------------------------------------------------*/
  /** Command line invocation of the neural network program.
   *  @param  args the command line arguments
   *  @since  2023.04.30 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public static void main (String[] args)
  { JMLP jmlp = new JMLP(); }   /* create MLP window */

}  /* class JMLP */
