/*----------------------------------------------------------------------
  File    : JMLPPanel.java
  Contents: Java panel to visualize multi-layer perceptron training
  Author  : Christian Borgelt
  History : 2023.05.17 file created from JMLP.java
----------------------------------------------------------------------*/
package jmlp;

import java.util.Random;
import java.awt.Dimension;
import java.awt.Color;
import java.awt.Graphics;
import javax.swing.JPanel;

/*--------------------------------------------------------------------*/
/** Class for the panel in which the neural network is drawn.
 *  @author Christian Borgelt
 *  @since  2023.04.30 */
/*--------------------------------------------------------------------*/
public class JMLPPanel extends JPanel {

  /*------------------------------------------------------------------*/
  /*  constants                                                       */
  /*------------------------------------------------------------------*/
  private static final double M_PI     =  3.14159265358979323846;
  private static final double MINVAL   = -0.2;  /* range of values */
  private static final double MAXVAL   =  1.2;  /* for diagrams */

  private static final int    xoff     =   0; /* x-offset for drawing */
  private static final int    yoff     = 399; /* y-offset for drawing */

  /** various colors for drawing */
  private static final Color  black    = Color.black;
  private static final Color  white    = Color.white;
  private static final Color  gray     = new Color(0xc0, 0xc0, 0xc0);
  private static final Color  graylite = new Color(0xe8, 0xe8, 0xe8);
  private static final Color  red      = Color.red;
  private static final Color  green    = new Color(0x00, 0xc0, 0x00);
  private static final Color  blue     = Color.blue;
  private static final Color  yellow   = new Color(0xff, 0xc0, 0x00);
  private static final Color  magenta  = Color.magenta;
  private static final Color  maglite  = new Color(0xff, 0xb0, 0xff);
  private static final Color  cyan     = new Color(0x00, 0xd0, 0xd0);
  private static final Color  cyanlite = new Color(0xb0, 0xd0, 0xd0);
  private static final Color  orange   = new Color(0xff, 0xa0, 0x00);
  private static final Color  oralite  = new Color(0xff, 0xd0, 0xb0);

  /* number of steps in x-direction for graph drawing */
  private static final int    N        = 52;

  /*------------------------------------------------------------------*/
  /*  instance variables                                              */
  /*------------------------------------------------------------------*/
  /** the random number generator */
  private Random  rand  = new Random();
  /** the initial weight range */
  private double  range = 1.0;
  /** the neural network / multi-layer perceptron */
  private MLP     mlp   = null;
  /** the outputs of the first  hidden neuron */
  private int[]   hd1   = new int[2*N+1];
  /** the outputs of the second hidden neuron */
  private int[]   hd2   = new int[2*N+1];
  /** the outputs of the output neuron */
  private int[]   out   = new int[2*N+1];

  /*------------------------------------------------------------------*/
  /** Create a multi-layer perceptron.
   *  @param  type  the problem type identifier
   *  @since  2023.04.30 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public JMLPPanel (int type)
  {                             /* --- create a panel for drawing */
    this.mlp = new MLP(type);   /* create and initialize an MLP */
    this.mlp.init(this.range, this.rand);
  }  /* JMPLPanel() */
  
  /*------------------------------------------------------------------*/
  /** Set the problem type.
   *  @param  type  the problem type identifier
   *  @since  2023.04.30 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void setType (int type)
  {                             /* --- set the problem type */
    this.mlp.setType(type);     /* set type and initialize MLP */
    this.mlp.init(this.range, this.rand);
    this.repaint();             /* repaint the panel */
  }  /* setType() */

  /*------------------------------------------------------------------*/
  /** Set the initial weight range.
   *  @param  range  the initial weight range
   *  @since  2023.04.30 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void setRange (double range)
  { this.range = range; }

  /*------------------------------------------------------------------*/
  /** Get the initial weight range.
   *  @return the initial weight range
   *  @since  2023.04.30 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public double getRange ()
  { return this.range; }

  /*------------------------------------------------------------------*/
  /** Get underlying multi-layer perceptron.
   *  @return the underlying multi-layer perceptron
   *  @since  2023.04.30 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public MLP getMLP ()
  { return this.mlp; }
  
  /*------------------------------------------------------------------*/
  /** Draw a (horizontal or vertical) arrow.
   *  @param  g   the graphics context for drawing
   *  @param  x1  the x-coordinate of the start point
   *  @param  y1  the y-coordinate of the start point
   *  @param  x2  the x-coordinate of the end   point
   *  @param  y2  the y-coordinate of the end   point
   *  @since  2023.04.30 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  private void arrow (Graphics g, int x1, int y1, int x2, int y2)
  {                             /* --- draw a horz. or vert. arrow */
    if (x1 == x2) {             /* if vertical arrow */
      g.drawLine(x1, y1, x2, y2+6);
      g.fillArc(x2-8, y2-8, 16, 16, 250, 40); }
    else {                      /* if horizontal arrow */
      g.drawLine(x1, y1, x2-6, y2);
      g.fillArc(x2-8, y2-8, 16, 16, 160, 40); }
  }  /* arrow() */

  /*------------------------------------------------------------------*/
  /** Draw a number / connection weight / bias value.
   *  @param  g      the graphics context for drawing
   *  @param  x      the x-coordinate of the text output
   *  @param  y      the y-coordinate of the text output
   *  @param  n      the number to draw
   *  @param  color  the color with which to draw the number
   *  @since  2023.04.30 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  private void number (Graphics g, int x, int y, double n, Color color)
  {                             /* --- print a number/weight */
    int    w;                   /* text width (in pixels) */
    String s;                   /* buffer for weight as a text */
 
    if ((n == -1.0) || (n == 0.0) || (n == 1.0))
      s = String.format(" %.2g ", n);
    else                        /* convert number to a string */
      s = String.format(" %.4g ", n);
    w = g.getFontMetrics().stringWidth(s);
    g.setColor(white);          /* clear background */
    g.fillRect(x-w/2, y-11, w, 13);
    g.setColor(color);          /* draw number/connection weight */
    g.drawString(s, x-w/2, y);
  }  /* number() */

  /*------------------------------------------------------------------*/
  /** Draw a (transformed) data point (logical function).
   *  @param  g      the graphics context for drawing
   *  @param  x      the x-coordinate of  the (transformed) data point
   *  @param  y      the y-coordinate of  the (transformed) data point
   *  @param  value  the output value for the (transformed) data point
   *  @param  size   the size of the point
   *  @param  color  the color with which to draw the point
   *  @since  2023.04.30 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  private void point (Graphics g, double x, double y, int value,
                      int diag, int size, Color color)
  {                             /* --- draw a (transformed) point */
    int wx = xoff +((diag > 0) ? 328 : 44);
    int wy = yoff -44;          /* window coordinates */
    wx += (int)(x *200 +0.5) -size/2;  /* compute window coordinates */
    wy -= (int)(y *200 +0.5) +size/2;  /* and draw point */
    g.setColor((value != 0) ? color : white);
    g.fillOval(wx+1, wy+1, size-2, size-2);
    g.setColor(color);
    g.drawOval(wx+1, wy+1, size-2, size-2);
    g.drawOval(wx,   wy,   size,   size);
}  /* point() */

  /*------------------------------------------------------------------*/
  /** Draw a(n inflection) line (or a parallel).
   *  @param  g       the graphics context for drawing
   *  @param  a       the coefficient of the x-coordinate
   *  @param  b       the coefficient of the y-coordinate
   *  @param  c       the offset / intercept term
   *  @param  diag    the diagram into which to draw
   *  @param  normal  whether to draw a normal vector
   *  @param  color   the color with which to draw the line
   *  @since  2023.04.30 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  private void line (Graphics g, double a, double b, double c,
                     int diag, boolean normal, Color color)
  {                             /* --- draw a (separating) line */
    int    wx = this.xoff +((diag > 0) ? 328 : 44);
    int    wy = this.yoff -44;  /* window coordinates */
    double x1, y1, x2, y2;      /* coordinates for lines */
    double t;                   /* temporary buffer */

    /* --- clipping --- */
    if (Math.abs(a) < 1e-12) a = (a < 0) ? -1e-12 : 1e-12;
    if (Math.abs(b) < 1e-12) b = (b < 0) ? -1e-12 : 1e-12;
    x1 = (-MINVAL *b -c) /a;
    if      (x1 < MINVAL) { x1 = MINVAL; y1 = (-MINVAL *a -c) /b; }
    else if (x1 > MAXVAL) { x1 = MAXVAL; y1 = (-MAXVAL *a -c) /b; }
    else                  {              y1 = MINVAL; }
    x2 = (-MAXVAL *b -c) /a;
    if      (x2 < MINVAL) { x2 = MINVAL; y2 = (-MINVAL *a -c) /b; }
    else if (x2 > MAXVAL) { x2 = MAXVAL; y2 = (-MAXVAL *a -c) /b; }
    else                  {              y2 = MAXVAL; }
    if ((y1 < MINVAL) || (y1 > MAXVAL)
    ||  (y2 < MINVAL) || (y2 > MAXVAL))
      return;                   /* skip invisible lines */

    /* --- drawing --- */
    g.setColor(color);          /* draw (inflection) line */
    g.drawLine(wx +(int)(x1 *200 +0.5), wy -(int)(y1 *200 +0.5),
               wx +(int)(x2 *200 +0.5), wy -(int)(y2 *200 +0.5));
    if (normal) {               /* if to draw a normal vector */
      t  = Math.hypot(a,b);     /* in the middle of the line */
      x1 = (x1 +x2) /2; x2 = x1 +0.07 *a/t;
      y1 = (y1 +y2) /2; y2 = y1 +0.07 *b/t;
      g.drawLine(wx +(int)(x1 *200 +0.5), wy -(int)(y1 *200 +0.5),
                 wx +(int)(x2 *200 +0.5), wy -(int)(y2 *200 +0.5));
      t = Math.atan2(y1-y2, x1-x2) *(180/M_PI);
      g.fillArc (wx +(int)(x2 *200 +0.5) -8, wy -(int)(y2 *200 +0.5)-8,
                 16, 16, (int)t-20, 40);
    }
  }  /* line() */

  /*------------------------------------------------------------------*/
  /** Draw training patterns for real-valued function.
   *  @param  g  the graphics context for drawing
   *  @since  2023.04.30 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  private void patterns (Graphics g)
  {                             /* --- draw training patterns */
    int x0 = this.xoff +572/2;  /* x-coordinate of origin */
    int y0 = this.yoff -28;     /* y-coordinate of origin */
    PatternSet patset;          /* pattern set */
    double[]   p;               /* to traverse the patterns */

    g.setColor(black);          /* get the number of patterns */
    patset = this.mlp.getPatternSet();  /* traverse the patterns */
    for (int n = patset.size(); --n >= 0; ) {
      p = patset.pattern(n);    /* get the next training pattern */
      g.fillRect(x0 +(int)(230 *p[0] +0.5)-1,
                 y0 -(int)(230 *p[2] +0.5)-1, 3, 3);
    }                           /* draw circles at the data points */
  }  /* patterns() */

  /*------------------------------------------------------------------*/
  /** Draw network output for real-valued function.
   *  @param  g  the graphics context for drawing
   *  @since  2023.04.30 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  private void netout (Graphics g)
  {                             /* --- draw network output */
    int    x0 = this.xoff +26;  /* x-coordinate of reference point */
    int    y0 = this.yoff -28;  /* y-coordinate of reference point */
    int    x;                   /* loop variable, pixel coordinate */

    for (x = 2*N+1; --x >= 0;){ /* traverse the input range */
      this.mlp.calc((x-N)*(5/230.0), 0);
      this.hd1[x] = y0 -(int)(230 *this.mlp.getOutput(0) +0.5);
      this.hd2[x] = y0 -(int)(230 *this.mlp.getOutput(1) +0.5);
      this.out[x] = y0 -(int)(230 *this.mlp.getOutput(2) +0.5);
    }                           /* compute outputs of neurons */
    g.setColor(blue);           /* output of first hidden neuron */
    for (x = 2*N+1; --x > 0; )
      g.drawLine(x0+5*x-5, hd1[x-1], x0+5*x, hd1[x]);
    g.setColor(green);           /* output of first hidden neuron */
    for (x = 2*N+1; --x > 0; )
      g.drawLine(x0+5*x-5, hd2[x-1], x0+5*x, hd2[x]);
    g.setColor(red);           /* output of output neuron */
    for (x = 2*N+1; --x > 0; )
      g.drawLine(x0+5*x-5, out[x-1], x0+5*x, out[x]);
  }  /* netout() */

  /*------------------------------------------------------------------*/
  /** Paint the whole panel (network and input/output spaces).
   *  @param  g  the graphics context for drawing
   *  @since  2023.04.30 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void paint (Graphics g)
  {                             /* --- (re)paint the whole panel */
    Dimension  d = getSize();   /* extensions of the panel */
    boolean    bool;            /* whether boolean function */
    int        x, y;            /* coordinates */
    double     w1, w2, theta;   /* weights and bias value */
    PatternSet patset;          /* pattern set */
    double[]   p;               /* individual pattern */
    double     shift;           /* shift value for bias */

    bool = (this.mlp.getType() < MLP.FUNC1);

    g.setColor(white);          /* draw white background */
    g.fillRect(0, 0, d.width, d.height);

    /* --- neurons --- */
    g.setColor(gray);           /* input neuron color */
    if (!bool) {                /* one input for real-valued function */
      g.fillOval(xoff+100, yoff-355, 22, 22);
      g.setColor(black);
      g.drawOval(xoff+100, yoff-355, 22, 22); }
    else {                      /* two inputs for Boolean function */
      g.fillOval(xoff+100, yoff-320, 22, 22);
      g.fillOval(xoff+100, yoff-390, 22, 22);
      g.setColor(black);
      g.drawOval(xoff+100, yoff-320, 22, 22);
      g.drawOval(xoff+100, yoff-390, 22, 22);
    }                           /* hidden and output neurons */
    g.setColor((bool) ? cyan    : blue);
    g.fillOval(xoff+275, yoff-320, 22, 22);
    g.setColor((bool) ? magenta : green);
    g.fillOval(xoff+275, yoff-390, 22, 22);
    g.setColor((bool) ? orange  : red);
    g.fillOval(xoff+450, yoff-355, 22, 22);
    g.setColor(black);          /* border of hidden & output neurons */
    g.drawOval(xoff+275, yoff-320, 22, 22);
    g.drawOval(xoff+275, yoff-390, 22, 22);
    g.drawOval(xoff+450, yoff-355, 22, 22);

    /* --- connections --- */
    g.setColor(black);          /* connection arrow color */
    if (bool) {                 /* two inputs for Boolean functions */
      arrow(g, xoff +40, yoff-379, xoff +97, yoff-379);    /* upper */
      arrow(g, xoff+111, yoff-372, xoff+111, yoff-387);    /* input */
      arrow(g, xoff +40, yoff-309, xoff +97, yoff-309);    /* lower */
      arrow(g, xoff+103, yoff-309, xoff+119, yoff-309);    /* input */
      /* upper hidden neuron: inputs and direction */
      g.drawLine(xoff+126, yoff-379, xoff+264, yoff-379);
      g.fillArc(xoff+266, yoff-380, 16, 16, 180, 40);
      arrow(g, xoff+126, yoff-379, xoff+272, yoff-379);
      arrow(g, xoff+286, yoff-372, xoff+286, yoff-387);
      /* lower hidden neuron: inputs and direction */
      g.drawLine(xoff+126, yoff-309, xoff+264, yoff-309);
      g.fillArc(xoff+266, yoff-324, 16, 16, 140, 42);
      arrow(g, xoff+126, yoff-309, xoff+272, yoff-309);
      arrow(g, xoff+278, yoff-309, xoff+294, yoff-309);
      g.drawLine(xoff+124, yoff-317, xoff+272, yoff-371);
      g.drawLine(xoff+124, yoff-371, xoff+272, yoff-317); }
    else {                      /* one input for Boolean functions */
      arrow(g, xoff +40, yoff-344, xoff +97, yoff-344);
      g.drawLine(xoff+126, yoff-348, xoff+271, yoff-379);
      g.fillArc(xoff+263, yoff-387, 16, 16, 172, 42);
      g.drawLine(xoff+126, yoff-340, xoff+271, yoff-309);
      g.fillArc(xoff+263, yoff-317, 16, 16, 146, 42);
    }                           /* connections to hidden neurons */
    /* connections to output neuron */
    g.drawLine(xoff+301, yoff-379, xoff+446, yoff-348);
    g.fillArc(xoff+438, yoff-356, 16, 16, 146, 42);
    g.drawLine(xoff+301, yoff-309, xoff+446, yoff-340);
    g.fillArc(xoff+438, yoff-348, 16, 16, 172, 42);
    /* arrow from output neuron */
    arrow(g, xoff+476, yoff-344, xoff+532, yoff-344);

    /* -- connection weights -- */
    if (bool) {                 /* if Boolean function */
      w1 = this.mlp.getWeight(MLP.HD1, MLP.IN1);
      number(g, xoff+198, yoff-304, w1, black);
      w2 = this.mlp.getWeight(MLP.HD1, MLP.IN2);
      number(g, xoff+158, yoff-353, w2, black);
      w1 = this.mlp.getWeight(MLP.HD2, MLP.IN1);
      number(g, xoff+158, yoff-325, w1, black);
      w2 = this.mlp.getWeight(MLP.HD2, MLP.IN2);
      number(g, xoff+198, yoff-374, w2, black); }
    else {                      /* if real-valued function */
      w1    = this.mlp.getWeight(MLP.HD1, MLP.IN1);
      theta = this.mlp.getWeight(MLP.HD1, MLP.BIAS);
      number(g, xoff+198, yoff-319, w1, black);
      number(g, xoff+110, yoff-299, -theta/w1, blue);
      w1    = this.mlp.getWeight(MLP.HD2, MLP.IN1);
      theta = this.mlp.getWeight(MLP.HD2, MLP.BIAS);
      number(g, xoff+198, yoff-359, w1, black);
      number(g, xoff+110, yoff-379, -theta/w1, green);
    }                           /* (connections to hidden neurons) */
    theta = -this.mlp.getWeight(MLP.HD1, MLP.BIAS);
    number(g, xoff+286, yoff-326, theta, black);
    theta = -this.mlp.getWeight(MLP.HD2, MLP.BIAS);
    number(g, xoff+286, yoff-352, theta, black);
    w1    =  this.mlp.getWeight(MLP.OUT, MLP.IN1);
    number(g, xoff+373, yoff-319, w1, black);
    w2    =  this.mlp.getWeight(MLP.OUT, MLP.IN2);
    number(g, xoff+373, yoff-359, w2, black);
    theta = -this.mlp.getWeight(MLP.OUT, MLP.BIAS);
    number(g, xoff+461, yoff-318, theta, black);

    /* --- separating line --- */
    g.setColor(gray);           /* separating line */
    g.drawLine( 0, yoff-289, d.width, yoff-289);

    if (bool) {                 /* if Boolean function */
      /* --- left diagram --- */
      g.setColor(gray);         /* auxiliary lines */
      g.drawLine(xoff+244, yoff -44, xoff+244, yoff-244);
      g.drawLine(xoff+ 44, yoff-244, xoff+244, yoff-244);
      g.setColor(black);        /* coordinate axes */
      arrow(g, xoff +44, yoff -44, xoff +44, yoff-264);
      arrow(g, xoff +44, yoff -44, xoff+265, yoff -44);
      patset = this.mlp.getPatternSet();
      p = patset.pattern(0);
      point(g, p[0], p[1], (int)p[2], 0, 8, yellow);
      p = patset.pattern(1);
      point(g, p[0], p[1], (int)p[2], 0, 8, green);
      p = patset.pattern(2);
      point(g, p[0], p[1], (int)p[2], 0, 8, blue);
      p = patset.pattern(3);
      point(g, p[0], p[1], (int)p[2], 0, 8, red);
      w1    = this.mlp.getWeight(MLP.HD1, MLP.IN1);
      w2    = this.mlp.getWeight(MLP.HD1, MLP.IN2);
      theta = this.mlp.getWeight(MLP.HD1, MLP.BIAS);
      line(g, w1, w2, theta, 0, true, cyan);
      for (int i = 1; i < 5; i += 1) {
        shift = Math.log((10-(double)i)/(double)i);
        line(g, w1, w2, theta-shift, 0, false, cyanlite);
        line(g, w1, w2, theta+shift, 0, false, cyanlite);
      }                         /* draw countours and 0.5 line */
      w1    = this.mlp.getWeight(MLP.HD2, MLP.IN1);
      w2    = this.mlp.getWeight(MLP.HD2, MLP.IN2);
      theta = this.mlp.getWeight(MLP.HD2, MLP.BIAS);
      line(g, w1, w2, theta, 0, true, magenta);
      for (int i = 1; i < 5; i += 1) {
        shift = Math.log((10-(double)i)/(double)i);
        line(g, w1, w2, theta-shift, 0, false, maglite);
        line(g, w1, w2, theta+shift, 0, false, maglite);
      }                         /* draw countours and 0.5 line */

      /* --- right diagram --- */
      g.setColor(graylite);
      for (int i = 348; i < 428; i += 20)
        g.drawLine(xoff+i, yoff -44, xoff+i, yoff-244);
      for (int i = 448; i < 528; i += 20)
        g.drawLine(xoff+i, yoff -44, xoff+i, yoff-244);
      for (int i =  64; i < 144; i += 20)
        g.drawLine(xoff+328, yoff-i, xoff+528, yoff-i);
      for (int i = 164; i < 244; i += 20)
        g.drawLine(xoff+328, yoff-i, xoff+528, yoff-i);
      g.setColor(gray);
      g.drawLine(xoff+428, yoff -44, xoff+428, yoff-244);
      g.drawLine(xoff+328, yoff-144, xoff+528, yoff-144);
      g.drawLine(xoff+528, yoff -44, xoff+528, yoff-244);
      g.drawLine(xoff+328, yoff-244, xoff+528, yoff-244);
      g.setColor(black);
      arrow(g, xoff+328, yoff -44, xoff+328, yoff-264);
      arrow(g, xoff+328, yoff -44, xoff+549, yoff -44);
      patset = this.mlp.getPatternSet();
      p = patset.pattern(0); this.mlp.calc(p[0], p[1]);
      point(g, this.mlp.getOutput(0), this.mlp.getOutput(1),
            (int)p[2], 1, 8, yellow);
      p = patset.pattern(1); this.mlp.calc(p[0], p[1]);
      point(g, this.mlp.getOutput(0), this.mlp.getOutput(1),
            (int)p[2], 1, 8, green);
      p = patset.pattern(2); this.mlp.calc(p[0], p[1]);
      point(g, this.mlp.getOutput(0), this.mlp.getOutput(1),
            (int)p[2], 1, 8, blue);
      p = patset.pattern(3); this.mlp.calc(p[0], p[1]);
      point(g, this.mlp.getOutput(0), this.mlp.getOutput(1),
            (int)p[2], 1, 8, red);
      w1    = this.mlp.getWeight(MLP.OUT, MLP.IN1);
      w2    = this.mlp.getWeight(MLP.OUT, MLP.IN2);
      theta = this.mlp.getWeight(MLP.OUT, MLP.BIAS);
      for (int i = 1; i < 5; i += 1) {
        shift = Math.log((10-(double)i)/(double)i);
        line(g, w1, w2, theta-shift, 1, false, oralite);
        line(g, w1, w2, theta+shift, 1, false, oralite);
      }                         /* draw countours and 0.5 line */
      line(g, w1, w2, theta, 1, true, orange); }

    else {                      /* if real-valued function */
      /* --- one diagram --- */
      g.setColor(gray);
      g.drawLine(xoff +20, yoff-258, xoff+552, yoff-258);
      g.drawLine(xoff +20, yoff-143, xoff+552, yoff-143);
      g.setColor(black);
      arrow(g, xoff+286, yoff -25, xoff+286, yoff-274);
      arrow(g, xoff+ 21, yoff -28, xoff+552, yoff -28);
      g.fillArc (xoff +12, yoff -36, 16, 16, -20, 40);
      g.drawLine(xoff+283, yoff-258, xoff+289, yoff-258);
      number(g, xoff+287, yoff-10, 0, black);
      g.drawLine(xoff +56, yoff -25, xoff +56, yoff -31);
      number(g, xoff +56, yoff-10, -1, black);
      g.drawLine(xoff+516, yoff -25, xoff+516, yoff -31);
      number(g, xoff+517, yoff-10, +1, black);
      patterns(g);              /* draw the training data and */
      netout(g);                /* the outputs of the neurons */
    }
  }  /* paint() */

}  /* class JMLPPanel */
